"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.awsConfig = void 0;
const ini_1 = __importDefault(require("ini"));
const debug_1 = __importDefault(require("debug"));
const paths_1 = require("./paths");
const mkdirp_1 = __importDefault(require("mkdirp"));
const fs_1 = __importDefault(require("fs"));
const util_1 = __importDefault(require("util"));
const debug = (0, debug_1.default)("aws-azure-login");
const writeFile = util_1.default.promisify(fs_1.default.writeFile);
// Autorefresh credential time limit in milliseconds
const refreshLimitInMs = 11 * 60 * 1000;
exports.awsConfig = {
    async setProfileConfigValuesAsync(profileName, values) {
        const sectionName = profileName === "default" ? "default" : `profile ${profileName}`;
        debug(`Setting config for profile '${profileName}' in section '${sectionName}'`);
        const config = (await this._loadAsync("config")) || {};
        config[sectionName] = {
            ...config[sectionName],
            ...values,
        };
        await this._saveAsync("config", config);
    },
    async getProfileConfigAsync(profileName) {
        const sectionName = profileName === "default" ? "default" : `profile ${profileName}`;
        debug(`Getting config for profile '${profileName}' in section '${sectionName}'`);
        const config = await this._loadAsync("config");
        if (!config) {
            return undefined;
        }
        return config[sectionName];
    },
    async isProfileAboutToExpireAsync(profileName) {
        debug(`Getting credentials for profile '${profileName}'`);
        const config = await this._loadAsync("credentials");
        let expirationDate;
        if (!config ||
            config[profileName] === undefined ||
            config[profileName].aws_expiration === undefined) {
            expirationDate = new Date();
        }
        else {
            expirationDate = new Date(config[profileName].aws_expiration);
        }
        const timeDifference = expirationDate.getTime() - new Date().getTime();
        debug(`Remaining time till credential expiration: ${timeDifference / 1000}s, refresh due if time lower than: ${refreshLimitInMs / 1000}s`);
        return timeDifference < refreshLimitInMs;
    },
    async setProfileCredentialsAsync(profileName, values) {
        const credentials = (await this._loadAsync("credentials")) || {};
        debug(`Setting credentials for profile '${profileName}'`);
        credentials[profileName] = values;
        await this._saveAsync("credentials", credentials);
    },
    async getAllProfileNames() {
        debug(`Getting all configured profiles from config.`);
        const config = (await this._loadAsync("config")) || {};
        const profiles = Object.keys(config).map(function (e) {
            return e.replace("profile ", "");
        });
        debug(`Received profiles: ${profiles.toString()}`);
        return profiles;
    },
    async _loadAsync(type) {
        if (!paths_1.paths[type])
            throw new Error(`Unknown config type: '${type}'`);
        return new Promise((resolve, reject) => {
            debug(`Loading '${type}' file at '${paths_1.paths[type]}'`);
            fs_1.default.readFile(paths_1.paths[type], "utf8", (err, data) => {
                if (err) {
                    if (err.code === "ENOENT") {
                        debug(`File not found. Returning undefined.`);
                        return resolve(undefined);
                    }
                    else {
                        return reject(err);
                    }
                }
                debug("Parsing data");
                // eslint-disable-next-line @typescript-eslint/no-explicit-any
                const parsedIni = ini_1.default.parse(data);
                // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
                return resolve(parsedIni);
            });
        });
    },
    async _saveAsync(type, data) {
        if (!paths_1.paths[type])
            throw new Error(`Unknown config type: '${type}'`);
        if (!data)
            throw new Error(`You must provide data for saving.`);
        debug(`Stringifying ${type} INI data`);
        const text = ini_1.default.stringify(data);
        debug(`Creating AWS config directory '${paths_1.paths.awsDir}' if not exists.`);
        await (0, mkdirp_1.default)(paths_1.paths.awsDir);
        debug(`Writing '${type}' INI to file '${paths_1.paths[type]}'`);
        await writeFile(paths_1.paths[type], text);
    },
};
